import math
import pyproj

import poobrains

from poobrains import app, flash, abort, locked_cached_property, Markup
from . import util


# Hacky but is only executed on boot and gives us all projections that don't need extra params
projection_choices = []
for proj_name, proj_title in pyproj.pj_list.items():
    try:
        pyproj.Proj(f"+proj={proj_name}")
    except pyproj.exceptions.CRSError:
        pass
    else:
        projection_choices.append((proj_name, proj_title))


@app.expose('/svg/plot')
class Plot(poobrains.svg.SVG):

    __supported_dtypes__ = []

    editor_fieldset = None # used to display a form to edit plot options
    editor_layer_fieldset = None # used to display a form to add/edit layers specific to this plot type

    padding = None
    width = None
    height = None
    plot_width = None
    plot_height = None
    description_height = None

    dataset = None
    length = None

    class Meta:

        modes = {
            'teaser': 'read',
            'full': 'read',
            'raw': 'read',
            'inline': 'read'
        }

    def __init__(self, handle=None, mode=None, dataset=None, layers=None, options=None, **kwargs):

        super().__init__(handle=handle, mode=mode, **kwargs)

        self._preprocessed = False

        if handle is None and dataset is None:
            abort(404, "No dataset selected")
        
        self.padding = app.config['SVG_PLOT_PADDING']
        self.plot_width = 100 - 2 * self.padding
        self.plot_height = self.plot_width
        self.description_height = app.config['SVG_PLOT_DESCRIPTION_HEIGHT']
        self.description_y = 100 - self.description_height
        self.width = self.plot_width + (2 * self.padding)
        self.height = self.plot_height + self.description_height + (3 * self.padding)

        if not dataset is None:
            self.dataset = dataset

        else:
            self.dataset = poobrains.analysis.data.load_dataset(handle)

        #if len(self.dataset) == 0:
        #    raise ValueError("Empty Dataset can't be visualized")

        if layers is None:
            for plotdef in self.dataset.plots:
                if plotdef.info['kind'] == self.__class__.__name__:
                    self.layers = plotdef.info['layers']
                    break

        else:
            self.layers = layers

        self.options = options

        self.preprocessed_data = None

    @property
    def length(self):
        return max([len(column['observations']) for column in self.dataset.values()])

    def layer_index(self, layer_name):
        return [k for k in self.layers.keys()].index(layer_name)

    def layer_id(self, layer_name):
        return f"dataset-{self.dataset.name}-{poobrains.helpers.clean_string(layer_name)}"

    def datapoint_id(self, layer_name, x):
        return f"dataset-{self.dataset.name}-{poobrains.helpers.clean_string(layer_name)}-{poobrains.helpers.clean_string(str(x))}"

    @locked_cached_property
    def palette(self):

        named_palette = {}
        palette = poobrains.svg.palettes.contrast_neighbors(poobrains.svg.palettes.get_highlight_color(), len(self.layers))

        for idx, (layer_name, layer) in enumerate(self.preprocessed_data.items()):
            if not layer['color'] is None:
                color = layer['color'].clone()
                color.alpha = palette[idx].alpha
            else:
                color = palette[idx]

            named_palette[layer_name] = color

        return named_palette

class XYFieldset(poobrains.formapp.MultistateFieldset):
    
    def __init__(self, plot_cls, parent, **kwargs):

        super().__init__(parent, **kwargs)
        self.plot_cls = plot_cls

    def build(self):

        super().build()

        scale_choices = [('linear', 'linear'), ('logarithmic', 'logarithmic')]

        self['scale_x'] = poobrains.form.fields.Select(
            label='Scale Y',
            choices=scale_choices,
            help_text="How to scale values on the X-axis"
        )

        self['scale_y'] = poobrains.form.fields.Select(
            label='Scale Y',
            choices=scale_choices,
            help_text="How to scale values on the X-axis"
        )

    def process(self, submit):
        flash("This needs to be implemented!")


class XYLayerFieldset(poobrains.formapp.MultistateFieldset):

    def __init__(self, plot_cls, parent, **kwargs):

        super().__init__(parent, **kwargs)

        self.plot_cls = plot_cls

    def build(self):

        super().build()
        visualizable_columns = self.formapp.dataset.columns_of_dtype(self.plot_cls.__supported_dtypes__)
        
        if 'layer_name' in self.parent.session:
            layer_name = self.parent.session['layer_name']
            layer_info = self.formapp.session['plot_definition_info']['layers'][layer_name]
        else:
            layer_name = None
            layer_info = {}
        
        column_choices = self.formapp.dataset.column_choices_of_dtype(self.plot_cls.__supported_dtypes__)
        description_choices = self.formapp.dataset.column_choices_of_dtype('object')

        self['layer_name'] = poobrains.form.fields.Text(
            label="Layer name",
            help_text='machine readable ([a-z], [0-9], -)',
            validators=[poobrains.form.validators.valid_name],
            default=layer_name
        )

        self['column_x'] = poobrains.form.fields.Select(
            label="X-axis column",
            choices=column_choices,
            help_text="Which column to use for the X-axis of the plot.",
            value=layer_info.get('x')
        )

        self['column_y'] = poobrains.form.fields.Select(
            label="Y-axis column",
            choices=column_choices,
            help_text="Which column to use for the Y-axis of the plot.",
            value=layer_info.get('y')
        )

        self['column_description'] = poobrains.form.fields.Select(
            label="Description column",
            choices=description_choices,
            help_text="Which column to take the observation/datapoint-level description from.",
            value=layer_info.get('description')
        )
      
        # TODO: I think this whole block can be deleted?
        if 'layer_name' in self.parent.session:
            self['layer_name'].value = self.parent.session['layer_name']
            
            if self['layer_name'].value in self.formapp.session['plot_definition_info']['layers']:
                self['column_x'].value = self.formapp.session['plot_definition_info']['layers'][self['layer_name'].value]['x']
                self['column_y'].value = self.formapp.session['plot_definition_info']['layers'][self['layer_name'].value]['y']

    def validate(self, submit):

        layer_name = self['layer_name'].value
        column_x = self['column_x'].value
        column_y = self['column_y'].value

        errors = poobrains.errors.CompoundError()

        if layer_name in (None, ''):
            errors.append(poobrains.errors.ValidationError("Layer name can't be empty!", self['layer_name']))

        if submit == 'add':
            if layer_name in self.formapp.session['plot_definition_info']['layers']:
                errors.append(poobrains.errors.ValidationError(f"Layer with name '{layer_name}' already exists!", self['layer_name']))

        elif submit == 'edit_apply':

            old_name = self.parent.session['layer_name']
            if layer_name != old_name and layer_name in self.formapp.session['plot_definition_info']['layers']:
                errors.append(poobrains.errors.ValidationError(f"Layer with name '{layer_name}' already exists!", self['layer_name']))

        if column_x is None:
            errors.append(poobrains.errors.ValidationError("You must choose an X-column for this layer.", self['column_x']))

        if column_y is None:
            errors.append(poobrains.errors.ValidationError("You must choose an Y-column for this layer.", self['column_y']))


        if len(errors):
            raise errors

    def process(self, submit):

        layer_name = self['layer_name'].value
        column_x = self['column_x'].value
        column_y = self['column_y'].value
        description = self['column_description'].value

        if submit == 'add':
                self.formapp.session['plot_definition_info']['layers'][layer_name] = {
                    'x': column_x,
                    'y': column_y,
                    'description': description
                }
                flash(f"Added layer '{layer_name}'.")


        elif submit == 'edit_apply':
            old_name = self.parent.session['layer_name']
            del(self.formapp.session['plot_definition_info']['layers'][old_name])
            self.formapp.session['plot_definition_info']['layers'][layer_name] = {
                'x': column_x,
                'y': column_y,
                'description': description
            }
            flash(f"Saved layer '{layer_name}'.")


@app.expose('/svg/scatterplot')
class ScatterPlot(Plot):
    __supported_dtypes__ = [
        'int32',
        'int64',
        'uint32',
        'uint64',
        'float32',
        'float64',
        'datetime64'
    ]

    layer_column_fields = ['x', 'y'] # keys in layer definitions that reference columns, needed in case of column renames.

    editor_fieldset = XYFieldset
    editor_layer_fieldset = XYLayerFieldset

    def __init__(self, handle=None, mode=None, dataset=None, layers=None, options=None, **kwargs):

        super().__init__(handle=handle, mode=mode, dataset=dataset, layers=layers, options=options, **kwargs)

        if self.layers is None:

            self.layers = {}
            if len(self.dataset) >= 2: # take first two columns as x and y axis, respectively
                iter_keys = iter(self.dataset.keys())
                x = next(iter_keys)
                y = next(iter_keys)
                self.layers[f'{x}-{y}'] = {
                    'x': x,
                    'y': y
                }

            else: # len(dataset) == 1; take only column as y with index as x; I don't think this is implemented yet…
                y = next(iter(self.dataset.keys()))
                self.layers[f'{y}'] = {
                    'y': y,
                }

    @property
    def is_valid(self):

        for layer_name, layer_info in self.layers.items():
            if not layer_info['x'] in self.dataset:
                return False
            if not layer_info['y'] in self.dataset:
                return False
        return True

    def preprocess_data(self):

        index = self.dataset.complete_index()

        preprocessed_data = {}
        #for column_name, column in self.dataset.items():
        for layer_name, layer_info in self.layers.items():

            column_x = self.dataset[layer_info['x']]
            column_y = self.dataset[layer_info['y']]

            dtype_x = util.__dtypes__[column_x['dtype']]
            dtype_y = util.__dtypes__[column_y['dtype']]

            point_description = False
            if layer_info['description'] in self.dataset:
                point_description = True
                point_descriptions = self.dataset[layer_info['description']]['observations']

            preprocessed_data[layer_name] = {
                'id': self.layer_id(layer_name),
                'title': column_y['title'],
                'description': poobrains.md.MarkdownString(column_y['description']), 
                'dtype': column_y['dtype'],
                'color': column_y['color'],
                'label_x': column_x['title'],
                'label_y': column_y['title'],
                'point_description': point_description,
                'points': [],
            }

            for idx in index:

                if idx in column_x['observations'] and idx in column_y['observations']:
                    x = column_x['observations'][idx]
                    y = column_y['observations'][idx]
                    
                    point = {
                        'id': self.datapoint_id(layer_name, x),
                        'x_raw': x,
                        'y_raw': y,
                        'x_pretty': dtype_x.numpy_pretty(x),
                        'y_pretty': dtype_y.numpy_pretty(y),
                        'x_float': dtype_x.numpy_vis(x),
                        'y_float': dtype_y.numpy_vis(y),
                    }

                    if point_description and idx in point_descriptions:
                        point['description'] = poobrains.md.MarkdownString(point_descriptions[idx])

                    preprocessed_data[layer_name]['points'].append(point)

        self.preprocessed_data = preprocessed_data

    def normalize_data(self):

        if self.preprocessed_data is None:
            raise RuntimeError("Called Plot.find_extreme before Plot.preprocess_data!")

        for preprocessed_layer in self.preprocessed_data.values():
            for point in preprocessed_layer['points']:
                point['x'] = self.normalize_x(point['x_float'])
                point['y'] = self.normalize_y(point['y_float'])

        self._preprocessed = True

    @locked_cached_property
    def min_x(self):
        return self.find_extreme('x', 'min')

    @locked_cached_property
    def max_x(self):
        return self.find_extreme('x', 'max')

    @locked_cached_property
    def min_y(self):
        return self.find_extreme('y', 'min')

    @locked_cached_property
    def max_y(self):
        return self.find_extreme('y', 'max')

    def find_extreme(self, axis, kind):

        if self.preprocessed_data is None:
            raise RuntimeError("Called Plot.find_extreme before Plot.preprocess_data!")

        key = f"{axis}_float"
        f = {'min': min, 'max': max}[kind]

        extremes = []
        #for series in self.dataset.values():
        for series in self.preprocessed_data.values():
            axis_values = [point[key] for point in series['points']]
            if len(axis_values) > 0:
                extremes.append(f(axis_values))

        if len(extremes) > 0:
            return f(extremes)
        return 0

    def normalize_x(self, value):

        if self.span_x == 0.0:
            return 50

        return (value - self.min_x) * (100 / self.span_x)

    def normalize_y(self, value):

        if self.span_y == 0.0:
            return 50

        return 100 - (value - self.min_y) * (100 / self.span_y)

    @locked_cached_property
    def labels(self):

        labels = {
            'x': {},
            'y': {}
        }

        for index, layer in self.preprocessed_data.items():
            for axis, key in {'x': 'label_x', 'y': 'label_y'}.items():
                if layer[key] in labels[axis]:
                    labels[axis][layer[key]].append(index)
                else:
                    labels[axis][layer[key]] = [index]

        return labels

    @locked_cached_property
    def label_x(self):

        #return u' / '.join([dataset.label_x for dataset in self.datasets])
        return self.dataset.label_x

    @locked_cached_property
    def label_y(self):

        #return u' / '.join([dataset.label_y for dataset in self.datasets])
        return self.dataset.label_y

    @locked_cached_property
    def has_grid(self):
        return self.span_x > 0 and self.span_y > 0

    @locked_cached_property
    def span_x(self):
        return self.max_x - self.min_x

    @locked_cached_property
    def span_y(self):
        return self.max_y - self.min_y

    @locked_cached_property
    def grid_x(self):

        if self.span_x == 0:
            return [{
                'value': self.min_x,
                'labels': [self.min_x]
            }]

        types = {self.dataset.data[layer['x']]['dtype'] for layer in self.layers.values()}
        grid_step = 10 ** (int(math.log10(self.span_x)) - 1) # wat?

        #offset = (self.min_x % grid_step) * grid_step # distance from start of plot to first line on the grid
        offset = 0
        start = self.min_x + offset

        x = start
        stops = []
        while x <= self.max_x:

            stop = {
                'value': x,
                'labels': []
            }

            for type in types:
                stop['labels'].append(util.__dtypes__[type].numpy_pretty(x))

            stops.append(stop)
            x += grid_step

        return stops

    @locked_cached_property
    def grid_y(self):

        if self.span_y == 0:
            return [self.min_y]

        grid_step = 10 ** (int(math.log10(self.span_y)) - 1)
        #offset = (self.min_y % grid_step) * grid_step # distance from start of plot to first line on the grid
        offset = grid_step - self.min_y % grid_step
        start = self.min_y + offset

        y = start
        coords = [y]
        while y <= self.max_y:
            coords.append(y)
            y += grid_step

        return coords

    def render(self, mode='teaser'):

        if mode in ('inline', 'raw') and not self._preprocessed:

            if not self.is_valid:
                return "Invalid plot definition."

            self.preprocess_data()
            self.normalize_data()

        return super().render(mode=mode)


@app.expose('/svg/lineplot')
class LinePlot(ScatterPlot):
    pass


@app.expose('/svg/barchart')
class BarChart(ScatterPlot):
    
    @locked_cached_property
    def bar_width(self):
        return (self.plot_width - self.padding) / self.length - 1


class MapFieldset(poobrains.formapp.MultistateFieldset):
    
    def __init__(self, plot_cls, parent, **kwargs):
       
        super().__init__(parent, **kwargs)

        self.plot_cls = plot_cls

    def build(self):

        super().build()

        if 'options' in self.formapp.session['plot_definition_info']:
            projection_default = self.formapp.session['plot_definition_info']['options'].get('projection')
        else:
            projection_default=None

        self['projection'] = poobrains.form.fields.Select(
            label="Map projection",
            choices=projection_choices,
            default=projection_default
        )

    def validate(self, submit):
        pass

    def process(self, submit):
        if submit == 'apply_options':
            self.formapp.session['plot_definition_info']['options'] = {
                'projection': self['projection'].value
            }


class MapLayerFieldset(poobrains.formapp.MultistateFieldset):

    def __init__(self, plot_cls, parent, **kwargs):

        super().__init__(parent, **kwargs)
        self.plot_cls = plot_cls

    def build(self):

        super().build()

        visualizable_columns = self.formapp.dataset.columns_of_dtype(self.plot_cls.__supported_dtypes__)

        geometry_column_choices = []
        for column_name, column in visualizable_columns.items():
            geometry_column_choices.append((column_name, column['title'] or column_name))

        chloroplethable_columns = self.formapp.dataset.columns_of_dtype(self.plot_cls.__chloropleth_dtypes__)
        chloropleth_column_choices = []
        for column_name, column in chloroplethable_columns.items():
            chloropleth_column_choices.append((column_name, column['title'] or column_name))
        
        description_choices = []
        for column_name, column in self.formapp.dataset.columns_of_dtype('object').items():
            description_choices.append((column_name, column['title'] or column_name))

        self['layer_name'] = poobrains.form.fields.Text(
            label="Layer name",
            help_text='machine readable ([a-z], [0-9], -)',
            validators=[poobrains.form.validators.valid_name],
        )

        self['column_geometry'] = poobrains.form.fields.Select(
            label="Geometry column",
            choices=geometry_column_choices,
            help_text="The colum which holds the geometry to be rendered."
        )

        self['column_chloropleth_a'] = poobrains.form.fields.Select(
            label="Chloropleth column",
            choices=chloropleth_column_choices,
            help_text="Select a column to be used to determine the coloring of map features."
        )

        self['column_chloropleth_b'] = poobrains.form.fields.Select(
            label="Chloropleth column (second)",
            choices=chloropleth_column_choices,
            help_text="Second column used to color map features, will lead to a bimorphic color palette."
        )

        self['column_description'] = poobrains.form.fields.Select(
            label="Description column",
            choices=description_choices,
            help_text="Which column to take the observation/datapoint-level description from.",
        )

        if 'layer_name' in self.parent.session:
            self['layer_name'].value = self.parent.session['layer_name']
            
            if self['layer_name'].value in self.formapp.session['plot_definition_info']['layers']:
                layer_name = self['layer_name'].value
                layer_info = self.formapp.session['plot_definition_info']['layers'][layer_name]
                self['column_geometry'].value = layer_info['geometry']
                self['column_chloropleth_a'].value = layer_info['chloropleth_a']
                self['column_chloropleth_b'].value = layer_info['chloropleth_b']
                self['column_description'].value = layer_info['description']

    def validate(self, submit):

        layer_name = self['layer_name'].value
        errors = poobrains.errors.CompoundError()

        if layer_name is None or layer_name == '':
            errors.append(poobrains.errors.ValidationError("Layer name can't be empty!", self['layer_name']))

        if submit == 'add':
            if layer_name in self.formapp.session['plot_definition_info']['layers']:
                errors.append(poobrains.errors.ValidationError(f"Layer with name '{layer_name}' already exists!", self['layer_name']))

        elif submit == 'edit_apply':

            old_name = self.parent.session['layer_name']
            if layer_name != old_name and layer_name in self.formapp.session['plot_definition_info']['layers']:
                errors.append(poobrains.errors.ValidationError(f"Layer with name '{layer_name}' already exists!", self['layer_name']))

        if len(errors):
            raise errors

    def process(self, submit):

        layer_name = self['layer_name'].value

        if submit == 'add':
            self.formapp.session['plot_definition_info']['layers'][layer_name] = {
                'geometry': self['column_geometry'].value,
                'chloropleth_a': self['column_chloropleth_a'].value,
                'chloropleth_b': self['column_chloropleth_b'].value,
                'description': self['column_description'].value
            }
            
        elif submit == 'edit_apply':
            old_name = self.parent.session['layer_name']
            del(self.formapp.session['plot_definition_info']['layers'][old_name])
            self.formapp.session['plot_definition_info']['layers'][layer_name] = {
                'geometry': self['column_geometry'].value,
                'chloropleth_a': self['column_chloropleth_a'].value,
                'chloropleth_b': self['column_chloropleth_b'].value,
                'description': self['column_description'].value
            }

@app.expose('/svg/map')
class Map(Plot):

    __supported_dtypes__ = ['geometry']
    __chloropleth_dtypes__ = [
        'int32',
        'int64',
        'uint32',
        'uint64',
        'float32',
        'float64',
        'datetime64'
    ]

    layer_column_fields = ['geometries', 'chloropleth_a', 'chloropleth_b'] # keys in layer definitions that reference columns, needed in case of column renames.
    editor_fieldset = MapFieldset
    editor_layer_fieldset = MapLayerFieldset

    def __init__(self, handle=None, mode=None, dataset=None, layers=None, options=None, **kwargs):

        super().__init__(handle=handle, mode=mode, dataset=dataset, layers=layers, **kwargs)
        
        if options is None or not 'projection' in options:
            projection = 'webmerc' # Web Pseudo Mercator. epsg:3857, ""WGS84""
        else:
            projection = options['projection']

        self.projection = pyproj.Proj(f"+proj={projection}")

        self.bbox = None
        self.bbox_projected = None
        self.reference_span_x = None
        self.reference_span_y = None
        self.aspect_ratio = None

    @property
    def is_valid(self):

        for layer_name, layer_info in self.layers.items():
            if not layer_info['geometry'] in self.dataset:
                return False

            for key in ('chloropleth_a', 'chloropleth_b'):
                if not layer_info[key] is None:
                    if not layer_info[key] in self.dataset:
                        return False

            return True

    def normalize_coord(self, coord):

        x, y = self.projection(coord[0], coord[1])

        if self.reference_span_x > 0:
            x -= self.bbox_projected[0][0]
            x /= self.reference_span_x
        else:
            x = 0.5

        if self.reference_span_y > 0:
            y -= self.bbox_projected[0][1]
            y /= self.reference_span_y
        else:
            y = -0.5

        x *= self.width
        y *= 100

        y *= -1 # y axes of geodata and SVG are inversely oriented
        return (x, y)

    def preprocess_data(self):

        chloropleth_palette_length = 5 # TODO: make this settable in the editor

        highlight_color = poobrains.svg.palettes.get_highlight_color()
        self.bbox = self.find_bbox()

        self.bbox_projected = [
            self.projection(self.bbox[0][0], self.bbox[0][1]),
            self.projection(self.bbox[1][0], self.bbox[1][1])
        ]

        self.reference_span_x = abs(self.bbox_projected[1][0] - self.bbox_projected[0][0])
        self.reference_span_y = abs(self.bbox_projected[0][1] - self.bbox_projected[1][1])

        if self.reference_span_x == 0 or self.reference_span_y == 0:
            self.aspect_ratio = 1
        else:
            self.aspect_ratio = self.reference_span_x / self.reference_span_y

        self.width = 100 * self.aspect_ratio

        self.preprocessed_data = {}

        base_palette = poobrains.svg.palettes.contrast_neighbors(highlight_color, len(self.layers))

        for i, (layer_name, layer_info) in enumerate(self.layers.items()):

            column_geometry = self.dataset[layer_info['geometry']]

            column_chloropleth_a = layer_info['chloropleth_a']
            column_chloropleth_b = layer_info['chloropleth_b']

            if not column_chloropleth_a is None:
                column_chloropleth_a = self.dataset[column_chloropleth_a]

            if not column_chloropleth_b is None:
                column_chloropleth_b = self.dataset[column_chloropleth_b]
            
            if not column_geometry['color'] is None:
                base_color = column_geometry['color']
            else:
                base_color = base_palette[i]

            if not column_chloropleth_a is None and not column_chloropleth_b is None: # 2D chloropleth

                title = f"{column_chloropleth_a['title']} vs. {column_chloropleth_b['title']}"
                description = "## {column_chloropleth_a['title']} ##\n{column_chloropleth_a['description']}\n\n\n## {column_chloropleth_b['title']} ##\n{column_chloropleth_b['description']}" # titled concatenation of both columns' descriptions
                if not column_chloropleth_a['color'] is None:
                    color_a = column_chloropleth_a['color']
                    base_color = color_a
                else:
                    color_a = base_color

                if not column_chloropleth_b['color'] is None:
                    color_b = column_chloropleth_b['color']
                else:
                    color_b = color_a.clone()
                    color_b.hue += 180

                palette = poobrains.svg.palettes.bipolar_color_progression(color_a, color_b, chloropleth_palette_length)
                chloropleth_axes = 2

                chloropleth_a_min = min(column_chloropleth_a['observations'])
                chloropleth_a_max = max(column_chloropleth_a['observations'])
                chloropleth_a_span = abs(chloropleth_a_max - chloropleth_a_min)
                chloropleth_a_step = chloropleth_a_span / chloropleth_palette_length

                chloropleth_b_min = min(column_chloropleth_b['observations'])
                chloropleth_b_max = max(column_chloropleth_b['observations'])
                chloropleth_b_span = abs(chloropleth_b_max - chloropleth_b_min)
                chloropleth_b_step = chloropleth_b_span / chloropleth_palette_length
                
                chloropleth_steps = []
                for i in range(0, chloropleth_palette_length):
                    raise RuntimeError("I should implement bipolar chloropleths.\n")

            elif not column_chloropleth_a is None: # 1D chloropleth

                title = column_chloropleth_a['title']
                description = column_chloropleth_a['description']
                if not column_chloropleth_a['color'] is None:
                    color = column_chloropleth_a['color']
                    base_color = color
                else:
                    color = base_color

                palette = poobrains.svg.palettes.single_value_progression(color, chloropleth_palette_length)
                chloropleth_axes = 1

                chloropleth_a_min = min(column_chloropleth_a['observations'].values())
                chloropleth_a_max = max(column_chloropleth_a['observations'].values())
                chloropleth_a_span = abs(chloropleth_a_max - chloropleth_a_min)
                chloropleth_a_step = chloropleth_a_span / chloropleth_palette_length

                chloropleth_steps = []
                for i in range(0, chloropleth_palette_length):
                    chloropleth_steps.append({
                        'axes': 1,
                        'min': chloropleth_a_min + chloropleth_a_step * i,
                        'max': chloropleth_a_min + chloropleth_a_step * (i + 1),
                        'color': palette[i],
                    })

            else: # no chloropleth
                title = column_geometry['title']
                description = column_geometry['description']
                palette = poobrains.svg.palettes.contrast_neighbors(base_color, len(column_geometry['observations']))
                chloropleth_axes = 0

            prepro_layer = {
                'id': self.layer_id(layer_name),
                'title': title,
                'description': poobrains.md.MarkdownString(description),
                'color': base_color,
                'palette': palette,
                'chloropleth_axes': chloropleth_axes,
                'geometries': [],
            }

            if chloropleth_axes > 0:
                prepro_layer['chloropleth_steps'] = chloropleth_steps

            for idx, feature in column_geometry['observations'].items():
                prepro_feature = self.preprocess_geometry(feature)
                prepro_layer['geometries'].append(prepro_feature)

            self.preprocessed_data[layer_name] = prepro_layer

        self._preprocessed = True

    def index_color(self, layer_name, idx):

        layer_info = self.layers[layer_name]
        prepro_layer = self.preprocessed_data[layer_name]
        if prepro_layer['chloropleth_axes'] == 2:
            column_a = self.dataset[layer_info['chloropleth_a']]
            column_b = self.dataset[layer_info['chloropleth_b']]
            value_a = column_a['observations'][idx]
            value_b = column_b['observations'][idx]

        elif prepro_layer['chloropleth_axes'] == 1:
            column = self.dataset[layer_info['chloropleth_a']]

            try:
                value = column['observations'][idx]
            except KeyError: # means there's no value for this observation in the chloropleth column, return transparent color
                return poobrains.svg.color.Color(red=0, green=0, blue=0, alpha=0)
            for step_info in prepro_layer['chloropleth_steps']:
                if value >= step_info['min'] and value <= step_info['max']:
                    color = step_info['color']
                    break
            else: # this else belongs to the for loop, *not* the if
                color = poobrains.svg.color.Color(red=255, green=0, blue=0, alpha=1)
        else:
            color = prepro_layer['palette'][idx % len(prepro_layer['palette'])]

        return color

    def merge_bboxes(self, a, b):

        return [
            [
                a[0][0] if a[0][0] < b[0][0] else b[0][0],
                a[0][1] if a[0][1] > b[0][1] else b[0][1],
            ],[
                a[1][0] if a[1][0] > b[1][0] else b[1][0],
                a[1][1] if a[1][1] < b[1][1] else b[1][1]
            ]
        ]

    def merge_bbox_with_point(self, bbox, coord):
        bbox = bbox.copy()
        if coord[0] < bbox[0][0]:
            bbox[0][0] = coord[0]

        if coord[0] > bbox[1][0]:
            bbox[1][0] = coord[0]

        if coord[1] < bbox[0][1]:
            bbox[0][1] = coord[1]

        if coord[1] > bbox[1][1]:
            bbox[1][1] = coord[1]

        return bbox

    def find_bbox(self):

        # lat/lon:
        #     * up/north is positive, down/south negative
        #     * left/west is negative, right/east positive

        # shapely .bounds:
        # minx, miny, maxx, maxy

        bbox_keyed = {
            'west': -180,
            'north': 80,
            'east': 180,
            'south': -80
        }
        #bbox = [[180, -80], [-180, 80]] <- i think this one is technically wrong

        for layer_name, layer_info in self.layers.items():

            column_geometry = self.dataset[layer_info['geometry']]
            for geometry in column_geometry['observations'].values():

                if geometry.bounds[0] < bbox_keyed['west']:
                    bbox_keyed['west'] = geometry.bounds[0]

                if geometry.bounds[1] < bbox_keyed['south']:
                    bbox_keyed['south'] = geometry.bounds[1]

                if geometry.bounds[2] > bbox_keyed['east']:
                    bbox_keyed['east'] = geometry.bounds[2]

                if geometry.bounds[3] > bbox_keyed['north']:
                    bbox_keyed['north'] = geometry.bounds[3]

        # basically points for top left and bottom right of the resulting map
        return [[bbox_keyed['west'], bbox_keyed['north']], [bbox_keyed['east'], bbox_keyed['south']]]

    def preprocess_geometry(self, geometry):

        prepro_geometry = {
            'type': geometry.type,
        }

        if geometry.type in ('GeometryCollection', 'MultiPoint', 'MultiLineString', 'MultiPolygon'):
            geometries = []
            for geometry in geometry:
                prepro_sub_geometry = self.preprocess_geometry(geometry)
                geometries.append(prepro_sub_geometry)

            prepro_geometry['geometries'] = geometries

        elif geometry.type in ('Point', 'LineString'):
            normalized_coords = []
            for coord in geometry.coords:
                normalized_coords.append(self.normalize_coord(coord))
            prepro_geometry['coords'] = normalized_coords

        elif geometry.type =='Polygon':
            exterior = []
            for coord in geometry.exterior.coords:
                exterior.append(self.normalize_coord(coord))
            prepro_geometry['exterior'] = exterior

            interiors = []
            for interior in geometry.interiors:
                normalized_interior = []
                for coord in interior.coords:
                    normalized_interior.append(self.normalize_coord(coord))
                interiors.append(normalized_interior)
            prepro_geometry['interiors'] = interiors

        return prepro_geometry

    def render_geometry(self, geometry):
        """
        """

        r = ""
        if geometry['type'] == 'Point':
            r += f'<use href="#marker" x="{geometry["coords"][0][0]}" y="{geometry["coords"][0][1]}" />'
        elif geometry['type'] == 'LineString':
            for idx in range(0, len(geometry['coords']) - 1):
                line_start = geometry['coords'][idx]
                line_end = geometry['coords'][idx + 1]
                r += f'<line x1="{line_start[0]}" y1="{line_start[1]}" x2="{line_end[0]}" y2="{line_end[0]}" />'
        elif geometry['type'] =='Polygon':
            r += f'<path d="M {geometry["exterior"][0][0]} {geometry["exterior"][0][1]}' # move "cursor" to first coordinate
            for coord in geometry['exterior'][1:-1]: # first and last coordinate are identical
                r += f' L {coord[0]} {coord[1]}' # make a line to the next coordinate
            r+= ' z ' # close linering

            if len(geometry['interiors']):
                for interior in geometry['interiors']:
                    r += f' M {interior[0][0]} {interior[0][1]}' # move cursor to first coordinate of interior ring
                    for coord in interior[1:-1]:
                        r += f' L {coord[0]} {coord[1]}' # make a line ot the next coordinate
                r+= ' z ' # close interior linering

            r += '" />'

        elif geometry['type'] in ('GeometryCollection', 'MultiPoint', 'MultiLineString', 'MultiPolygon'):
            for sub_geometry in geometry['geometries']:
                r += self.render_geometry(sub_geometry)

        return Markup(r)

    def render(self, mode='teaser'):

        if mode in ('inline', 'raw') and not self._preprocessed:

            if not self.is_valid:
                return "Invalid plot definition."

            self.preprocess_data()

        return super().render(mode=mode)
