"""
The documentation system.
"""

#import __builtin__

import os
import re
import sys
import string
import codecs # needed to open files with utf-8 encoding
import collections
import inspect
import pkgutil
import pydoc
import jinja2

from poobrains import app, Response
import poobrains.errors
import poobrains.helpers
import poobrains.storage
import poobrains.auth
import poobrains.md


def clean(text):

    """ mainly for cleaning < and > from `Repr`s """

    return text.replace('<', '&lt;').replace('>', '&gt;')


class MDRepr(pydoc.TextRepr, object):

    def repr_string(self, x, level):
        return clean(super(MDRepr, self).repr_string(x, level))

    def repr1(self, x, level):
        return clean(super(MDRepr, self).repr1(x, level))


class PooDoc(pydoc.HTMLDoc, object):

    """
    pydoc-compatible HTML5 renderer
    """


    level_offset = None
    local = None
    index_url = None

    def __init__(self, level_offset=1, local=False, index_url='/doc/'): 

        self.level_offset = level_offset
        self.local = local
        self.index_url = index_url

    # utility functions
    
    def url(self, name):

        """ Create URL for a documentable thing. Mainly intended for subclassing """

        return f"/doc/{name}/"

    def url_object(self, object, parent=None, name=None):

        if name is None:
            name = object.__name__

        if inspect.ismodule(object):
            url = self.url(object.__name__)

        #elif inspect.ismethod(object):
        elif inspect.isroutine(object) and inspect.isclass(parent):
            url = self.url(parent.__module__)
            url += f'#{parent.__name__}.{name}' # qualname is classname.methodname

        elif pydoc.isdata(object):
            if inspect.ismodule(parent):
                url = self.url(parent.__name__)
                url += f'#{name}'
            else:
                url = self.url(parent.__module__)
                url += f'#{parent.__name__}.{name}'

        #elif inspect.isclass(object):
        else:

            url = self.url(object.__module__)
            url += f'#{object.__name__}'

        return url

    def link(self, url, label):
        return f'<a href="{url}">{label}</a>'

    def link_external(self, url, label):
        return f'<a href="{url}" target="_blank" rel="noreferrer noopener">{label}</a>'

    def link_object(self, object, parent=None, name=None, label=None):

        if name is None:
            name = object.__name__

        url = self.url_object(object, parent=parent, name=name)

        if label is None:
            label = object.__name__

        return self.link(url, label)

    def heading(self, level, content, css_class=None):

        """ Create a HTML heading """

        if level > 6:
            level = 6 # HTML headings only go from h1 to h6

        if not css_class is None:
            return f'<h{level} class="{css_class}">{content}</h{level}>'

        return f'<h{level}>{content}</h{level}>'

    # documentation logic

    def mrotree(self, cls):
        result = '<ul class="mro">'
        for base in cls.__bases__:
            result += f'<li>{self.link_object(base, parent=cls)}{self.mrotree(base)}</li>'
        result += '</ul>'

        return result

    def document(self, object, name=None, *args):

        """Generate documentation for an object."""

        args = (object, name) + args
        kwargs = {'level': self.level_offset + 1}
        # 'try' clause is to attempt to handle the possibility that inspect
        # identifies something in a way that pydoc itself has issues handling;
        # think 'super' and how it is a descriptor (which raises the exception
        # by lacking a __name__ attribute) and an instance.
        if inspect.isgetsetdescriptor(object): return self.docdata(*args, **kwargs)
        if inspect.ismemberdescriptor(object): return self.docdata(*args, **kwargs)
        try:
            if inspect.ismodule(object): return self.docmodule(*args, **kwargs)
            if inspect.isclass(object): return self.docclass(*args, **kwargs)
            if inspect.isroutine(object): return self.docroutine(*args, **kwargs)
        except AttributeError:
            pass
        if isinstance(object, property): return self.docproperty(*args, **kwargs)
        return self.docother(*args, **kwargs)

    def docmodule(self, object, name=None, mod=None, *ignored, level=1):

        if name is None:
            name = object.__name__
        
        try:
            all = object.__all__
        except AttributeError:
            all = None

        components = {}

        # build the head
        parts = name.split('.')
        links = []

        for i in range(len(parts)-1):
            url = self.url('.'.join(parts[:i+1]))
            links.append(f'<a href="{url}">{parts[i]}</a>')

        head_link = '.'.join(links + parts[-1:])
        components['heading'] = self.heading(level, head_link)

        # build link to official docs, if any
        docloc = self.getdocloc(object)
        if not docloc is None:
            components['docloc'] = self.link_external(docloc, 'Official documentation')

        if hasattr(object, '__author__'):
            components['author'] = self.heading(level + 1, 'Author') + f'<p>{object.__author__}</p>'

        if hasattr(object, '__credits__'):
            components['credits'] = self.heading(level + 1, 'Credits') + f'<p>{object.__credits__}</p>'

        # build the main description (i.e. rendering .__doc__)
        if not object.__doc__ is None:
            components['description'] = poobrains.md.md.convert(poobrains.helpers.unindent_string(object.__doc__))
  
        # build list of submodules if object is a package
        if hasattr(object, '__path__'): # means this is a package
            mod_links = []
            for importer, mod_name, ispkg in pkgutil.iter_modules(object.__path__):
                mod_domain = '.'.join((name, mod_name))
                url = self.url(mod_domain)
                mod_links.append(self.link(url, mod_name))

            if len(mod_links):
                components['submodules'] = self.heading(level + 1, 'Submodules')
                components['submodules'] += '<ul class="submodules">'
                components['submodules'] += '\n'.join([f'<li>{modlink}</li>' for modlink in mod_links])
                components['submodules'] += '</ul>'

        # build docs for any classes in this module
        classes = {}
        for name, cls in inspect.getmembers(object, inspect.isclass):
            if (inspect.getmodule(cls) or object) is object and pydoc.visiblename(name, all, object):
                classes[name] = cls

        if len(classes):
            components['classes'] = self.heading(level + 1, 'Classes')
            components['classes'] += '<ul class="doc-classes">'
            for name, cls in classes.items():
                components['classes'] += '<li>'
                components['classes'] += self.docclass(cls, name=name, mod=object, level=level+2)
                components['classes'] += '</li>'
            components['classes'] += '</ul>'

        # build docs for any routines/functions in this module
        routines = {}
        for name, routine in inspect.getmembers(object, inspect.isroutine):
            if pydoc.visiblename(name, all, object):
                routines[name] = routine

        if len(routines):
            components['routines'] = self.heading(level + 1, 'Routines')
            components['routines'] += '<ul class="doc-routines">'
            for name, routine in routines.items():
                components['routines'] += '<li>'
                components['routines'] += self.docroutine(routine, name=name, mod=object, level=level+2)
                components['routines'] += '</li>'
            components['routines'] += '</ul>'

        # build docs for any pieces of data in this module
        data = {}
        for name, datum in inspect.getmembers(object, pydoc.isdata):
            if pydoc.visiblename(name, all, object):
                data[name] = datum

        if len(data):
            components['data'] = self.heading(level + 1, 'Data')
            components['data'] += '<ul class="doc-data">'
            for name, datum in data.items():
                components['data'] += '<li>'
                components['data'] += self.docdata(datum, name=name, mod=object, level=level+2)
                components['data'] += '</li>'
            components['data'] += '</ul>'

        # stitch together the output
        result = '<article id="doc-top" class="doc-module">'

        result += components['heading']

        if 'docloc' in components:
            result += components['docloc']

        if 'author' in components:
            result += components['author']

        if 'credits' in components:
            result += components['credits']

        if 'description' in components:
            result += components['description']

        if 'submodules' in components:
            result += components['submodules']

        if 'classes' in components:
            result += components['classes']

        if 'routines' in components:
            result += components['routines']

        if 'data' in components:
            result += components['data']

        result += '</article>'

        return result

    def docclass(self, object, name=None, mod=None, funcs={}, classes={}, *ignored, level=1):

        try:
            all = mod.__all__
        except AttributeError:
            all = None

        components = {}

        components['heading'] = self.link_object(object, parent=mod, label=self.heading(level, f'class {object.__name__}'))

        # build method resolution order
        mro = inspect.getmro(object)
        if len(mro) > 2:
            components['mro'] = '<details><summary>Method Resolution Order</summary>'
            components['mro'] += self.mrotree(object)
            components['mro'] += '</details>'

        if not object.__doc__ is None:
            components['description'] = poobrains.md.md.convert(poobrains.helpers.unindent_string(object.__doc__))

        # collect and document methods
        methods = {}
        #for name, method in inspect.getmembers(object, inspect.ismethod):
        # inspect.ismethod only returns functions already bound.
        # i.e. given a class, it *won't* find its methods, but only
        # when given an *object* of that class.
        for name, method in inspect.getmembers(object, inspect.isroutine):
            if pydoc.visiblename(name, all, mod):
                methods[name] = method

        if len(methods):
            components['methods'] = self.heading(level + 1, 'Methods')
            components['methods'] += '<ul class="doc-methods">'
            for name, method in methods.items():
                components['methods'] += '<li>'
                components['methods'] += self.docroutine(method, name=name, mod=mod, cl=object, level=level+2)
                components['methods'] += '</li>'
            components['methods'] += '</ul>'

        # collect and document data
        data = {}
        for name, datum in inspect.getmembers(object, pydoc.isdata):
            if pydoc.visiblename(name, all, mod):
                data[name] = datum

        if len(data):
            components['data'] = self.heading(level + 1, 'Data')
            components['data'] += '<ul class="doc-data">'
            for name, datum in data.items():
                components['data'] += '<li>'
                components['data'] += self.docdata(datum, name=name, mod=mod, cl=object, level=level+2)
                components['data'] += '</li>'
            components['data'] += '</ul>'

        # stitch together the result
        result = f'<article id="{object.__qualname__}" class="doc-class">'
        result += components['heading']

        if 'mro' in components:
            result += components['mro']

        if 'description' in components:
            result += components['description']

        if 'methods' in components:
            result += components['methods']

        if 'data' in components:
            result += components['data']

        result += '</article>'

        return result

    def docroutine(self, object, name=None, mod=None, funcs={}, classes={}, methods={}, cl=None, level=1):

        if name is None:
            name = object.__name__

        components = {}
        
        components['heading'] = self.link_object(object, parent=cl or mod, name=name, label=self.heading(level, name))

        # find module / class this routine was originally defined in

        mod_obj = inspect.getmodule(object)
        original_reference_unknowable = False

        if mod_obj is None: # cover edge cases for when inspect.getmodule fails
            cl_name_obj, *_ = object.__qualname__.split('.') # name of the class actually defining this function
            if cl_name_obj == cl.__name__:
                mod_obj = mod
            elif hasattr(sys.modules['builtins'], cl_name_obj):
                mod_obj = sys.modules['builtins']
            else:
                components['original_reference'] = f'<aside>Can\'t figure out {object.__qualname__} – might be magic!</aside>'

        if not 'original_reference' in components:

            if not cl is None: # means this routine is part of a class
                link = False
                if hasattr(object, '__qualname__'): # classes implementing __call__ are routines but have no __qualname__ (as of 3.7)
                    cl_name_obj, *_ = object.__qualname__.split('.') # name of the class actually defining this function
                else:
                    # object is an instantiated object of a class implementing __call__ (or similar shitfuckery)
                    # in this case, we can't really do better than assume we're in the right class.
                    obj_wrapper = type(object)
                    obj_wrapper_mod = inspect.getmodule(obj_wrapper)
                    obj_wrapper_link = self.link_object(obj_wrapper, parent=obj_wrapper_mod)
                    components['wrapper'] = f'<aside class="doc-routine-wrapper">This routine is a special sally of type {obj_wrapper_link}</aside>'

                    if hasattr(object, 'func'): # handles flask.helpers.locked_cached_property (and hopefully some others)
                        object = object.func # do the rest of the docroutine logic on the actual function
                        mod_obj = inspect.getmodule(object) # not handling builtins here, just assuming getmodule will work in these cases
                        cl_name_obj, *_ = object.__qualname__.split('.') # name of the class actually defining this function
                    else: # can't get original function and thus not module and class its defined in
                        original_reference_unknowable = True

                if original_reference_unknowable:
                    components['original_reference'] = '<aside class="doc-error">Can\'t figure out where this routine was initially defined!</aside>'

                else:

                    if mod != mod_obj: # function defined in different class in a different module
                        if mod_obj is sys.modules['builtins']:
                            label = f'{cl_name_obj}.{name}' # no module name for builtins
                        else:
                            label = f'{mod_obj.__name__}.{cl_name_obj}.{name}'
                        cl_obj = getattr(mod_obj, cl_name_obj)
                        link = self.link_object(object, parent=cl_obj, name=name, label=label)
                    elif cl.__name__ != cl_name_obj: # function defined in different class in the same module
                        cl_obj = getattr(mod, cl_name_obj)
                        link = self.link_object(object, parent=cl_obj, name=name, label=f'{cl_name_obj}.{name}')

                    if link != False:
                        components['original_reference'] = f'<aside>This function is inherited from {link}.</aside>'

            #elif not '.' in object.__qualname__: # means this function is a direct member of a module, i.e. not in a class
            else: # means this function is a direct member of a module, i.e. not in a class
                if mod != mod_obj:
                    link = self.link_object(object, parent=mod_obj, name=name, label=f'{mod_obj.__name__}.{name}')
                    components['original_reference'] = f'<aside>This function is imported from {link}.</aside>'

        if not 'original_reference' in components: # skip all other documentation if this function is only imported/inherited

            # build highlighted source code in <details>, if possible
            try: 
                code = inspect.getsource(object)
                code = poobrains.helpers.unindent_string(code)
                code = "``` python\n" + code + "\n```" # wrap source code in markdown fenced code block
            except TypeError:
                pass
            else:
                components['code'] = '<details class="doc-code"><summary>Source</summary>'
                components['code'] += poobrains.md.md.convert(code)
                components['code'] += '</details>'

            # build the main description (i.e. rendering .__doc__)
            if not object.__doc__ is None:
                components['description'] = poobrains.md.md.convert(poobrains.helpers.unindent_string(object.__doc__))

            # build list of parameters
            try:
                argspec = inspect.getfullargspec(object)
            except TypeError as e:
                components['args'] = '<aside class="doc-error error">Could not detect parameters!</aside>'

            else:

                args = {
                    'positional': [],
                    'keyword': [],
                    'count': 0
                }

                # collect info on positional args
                if not argspec.defaults is None:
                    defaults_len = len(argspec.defaults)
                else:
                    defaults_len = 0
                defaults_offset = len(argspec.args) - defaults_len


                for idx, arg_name in enumerate(argspec.args):
                    arg_info = {'name': arg_name}
                    if defaults_len > 0 and idx >= defaults_offset:
                        arg_info['default'] = argspec.defaults[idx - defaults_offset]
                    args['positional'].append(arg_info)
                    args['count'] += 1

                if not argspec.varargs is None:
                    args['count'] += 1

                # collect info on keyword args
                for idx, arg_name in enumerate(argspec.kwonlyargs):
                    arg_info = {'name': arg_name}
                    arg_info['default'] = argspec.kwonlydefaults[arg_name]
                    args['keyword'].append(arg_info)
                    args['count'] += 1

                if not argspec.varkw is None:
                    args['count'] += 1

                if args['count'] > 0:

                    components['args'] = self.heading(level + 1, "Parameters")
                    components['args'] += '<ul class="doc-routine-args">'

                    for arg_info in args['positional']:
                        if 'default' in arg_info:
                            components['args'] += '<li class="doc-arg positional keyword">'
                        else:
                            components['args'] += '<li class="doc-arg positional">'
                        components['args'] += f'<span class="doc-arg-name">{arg_info["name"]}</span>'
                        if 'default' in arg_info:
                            default_type = type(arg_info['default'])
                            default_type_link = self.link_object(default_type)
                            components['args'] += f' = <span class="doc-arg-default">{repr(arg_info["default"])}</span> (<span class="doc-arg-default-type">{default_type_link}</span>)'
                        components['args'] += '</li>'

                    if argspec.varargs:
                        components['args'] += f'<li><span class="doc-arg varargs">*{argspec.varargs}</span></li>'

                    for arg_info in args['keyword']:
                        components['args'] += '<li class="doc-arg keyword keyword-only">'
                        components['args'] += f'<span class="doc-arg-name">{arg_info["name"]}</span>'
                        default_type = type(arg_info['default'])
                        default_type_link = self.link_object(default_type)
                        components['args'] += f' = <span class="doc-arg-default">{repr(arg_info["default"])}</span> (<span class="doc-arg-default-type">{default_type_link}</span>)'
                        components['args'] += '</li>'

                    components['args'] += '</ul>'

        if cl:
            result = f'<article id="{cl.__name__}.{name}" class="doc-routine">'
        else:
            result = f'<article id="{name}" class="doc-routine">'

        result += components['heading']

        if 'wrapper' in components:
            result += components['wrapper']

        if 'original_reference' in components:
            result += components['original_reference']

        if 'code' in components:
            result += components['code']

        if 'description' in components:
            result += components['description']

        if 'args' in components:
            result += components['args']

        result += '</article>'

        return result
    
    def docdata(self, object, name=None, mod=None, cl=None, level=1):

        if name is None:
            name = object.__name__

        components = {}

        components['heading'] = self.link_object(object, parent=cl or mod, name=name, label=self.heading(level, name, css_class="doc-datum-name"))

        obj_type = type(object)
        obj_type_link = self.link_object(obj_type, parent=inspect.getmodule(obj_type))
        components['value'] = f' = <span class="doc-datum-value">{repr(object)}</span>(<span class="doc-datum-type">{obj_type_link}</span>)'

        if cl:
            result = f'<article id="{cl.__name__}.{name}" class="doc-datum">'
        else:
            result = f'<article id="{name}" class="doc-datum">'

        result += components['heading']
        result += components['value']

        result += '</article>'

        return result

    def docother(self, object, name=None, mod=None, parent=None, maxlen=None, doc=None, level=1):

        if name is None:
            name = object.__name__

        components = {}

        components['heading'] = self.link_object(object, parent=mod, name=name, label=self.heading(level, name))

        if hasattr(mod, '__name__'):
            result = f'<article id="{mod.__name__}.{name}" class="doc-other">'
        else:
            result = '<article class="doc-other">'

        result += components['heading']

        result += "This is something else!"

        result += '</article>'

        return result


@app.expose('/doc/')
class Documentation(poobrains.auth.Protected):

    """
    A `Renderable` that can create and render module documentation on the fly.
    """

    title = None
    text = None

    def view(self, mode='full', **kwargs):

        md_dir = os.path.join(app.poobrain_path, 'doc')
        md_path = os.path.join(md_dir, f'{self.handle_string}.md')

        if self.handle_string is None: # documentation frontpage

            text = ''
            #for modname in sys.modules:
            #    if not '.' in modname and not modname.startswith('_'):
            #        text += f'* [documentation/{modname}]\n'

            text += '## Descriptive documents ##\n'
            for filename in os.listdir(md_dir):
                if filename.split('.')[-1] == 'md': # means we have a .md file
                    docname = filename[0:-3] # filename without the ending '.md'
                    text += f'* [{docname}][documentation/{docname}]\n'

            text += '\n## Reference ##\n'
            text += '* [poobrains][documentation/poobrains]\n'
            
            self.title = 'Documentation'
            self.text = poobrains.md.MarkdownString(text)

        elif os.path.exists(md_path):

            self.title = self.handle_string.title()
            self.text = poobrains.md.MarkdownString(codecs.open(md_path, 'r', encoding='utf-8').read())

        elif self.handle_string in sys.modules:
            
            doc = PooDoc()
            subject = sys.modules[self.handle_string]
            self.title = subject.__name__
            self.text = poobrains.rendering.RenderString(jinja2.Markup(doc.document(subject)))

        else: # module that doesn't exist or isn't loaded
            if self.handle_string.startswith('poobrains.'):
                import importlib
                try:
                    subject = importlib.import_module(self.handle_string)
                except ModuleNotFoundError:
                    raise poobrains.errors.ExposedError("Unknown module.")
                else:
                    doc = PooDoc()
                    self.title = subject.__name__
                    self.text = poobrains.rendering.RenderString(jinja2.Markup(doc.document(subject)))
                    del(subject) # unload module again
            else:
                raise poobrains.errors.ExposedError("Unloaded or unknown module that's not part of poobrains. Not loading for security reasons.")


        rv, status = super(Documentation, self).view(mode=mode, **kwargs)
        r = Response(rv, status)
        r.cache_control.public = True
        r.cache_control.max_age = app.config['CACHE_LONG']

        return r
