import collections
import peewee
import jinja2
import markdown

import flask

# local imports
from poobrains import app
import poobrains.rendering
import poobrains.storage
#import poobrains.auth


def magic_markdown_loader(storable, handle):

    storables = poobrains.storage.Storable.class_children_keyed(lower=True)

    if storable.lower() in storables:
        cls = storables[storable.lower()]
        return cls.load(handle)

    else:
        renderables = poobrains.rendering.Renderable.class_children_keyed(lower=True)
        cls = renderables[storable.lower()]
        return cls(handle=handle) # we could try handling Renderables without handle, but I think it's needed for expose anyways.

    return False


class MarkdownString(collections.UserString):

    def __init__(self, seq='', css_class=''):

        self._name = None
        self.css_class = css_class
        super(MarkdownString, self).__init__(seq)


    def __get__(self, instance, cls):
        if instance is None:
            return self

        else:
            return instance.__dict__[self._name]


    def __set__(self, instance, value):
        instance.__dict__[self._name] = type(self)(value)


    def __set_name__(self, owner, name):
        self._name = name


    def __add__(self, other):
        return type(self)(str(self) + str(other))


    def __radd__(self, other):
        return self.__add__(other)


    def __iadd__(self, other):
        return self.__add__(other)


    def render(self, mode='inline'): # mode is ignored anyways
        html = md.convert(self)
        return jinja2.Markup(f'<div class="markdown {self.css_class}">{html}</div>')


class MarkdownFieldAccessor(peewee.FieldAccessor):

    def __set__(self, instance, value):

        if not isinstance(value, MarkdownString):
            value = MarkdownString(value)

        super(MarkdownFieldAccessor, self).__set__(instance, value)


class MarkdownField(poobrains.storage.fields.TextField):

    accessor_class = MarkdownFieldAccessor

poobrains.storage.fields.MarkdownField = MarkdownField


class DisplayRenderable(markdown.inlinepatterns.Pattern):

    def handleMatch(self, match):

        if match:

            cls_name = match.group(2).lower()
            handle = match.group(3)

            renderables = poobrains.rendering.Renderable.class_children_keyed(lower=True)

            if cls_name in renderables:

                cls = renderables[cls_name.lower()]
                try:

                    if issubclass(cls, poobrains.storage.Storable):
                        instance = cls.load(handle)

                    else:
                        instance = cls(handle=handle)

                    #if isinstance(instance, poobrains.auth.Protected):
                    if hasattr(instance, 'permissions'):

                        try:
                            instance.permissions['read'].check(flask.g.user)
                        
                        except:
                            return jinja2.Markup(md.convert(f"*You are not allowed to view an instance of {cls.__name__} that was placed here.*"))
                    if 'inline' in instance._meta.modes:
                        return instance.render('inline')
                    elif 'teaser' in instance._meta.modes:
                        return instance.render('teaser')
                    elif 'full' in instance._meta.modes:
                        return instance.render('full')
                    else:
                        return instance.render()

                except Exception as e:
                    app.logger.debug(e)
                    return jinja2.Markup(f'<span class="markdown-error">{cls.__name__} could not be loaded.</span>')

            else:
                return jinja2.Markup(f"<span class=\"markdown-error\">Don't know what {cls.__name__} is.</span>")

        return super(DisplayRenderable, self).handleMatch(match)


class DisplayRenderableExtension(markdown.Extension):

    def extendMarkdown(self, md, md_globals):

        md.inlinePatterns.add(
            self.__class__.__name__,
            DisplayRenderable('\!\[(.*?)/(.*?)]'),
            '<reference'
        )

extra_extensions = app.config['MARKDOWN_EXTENSIONS']
if 'markdown.extensions.codehilite' in extra_extensions:
    import markdown.extensions.codehilite
    extra_extensions.pop(extra_extensions.index('markdown.extensions.codehilite'))
    extra_extensions.append(markdown.extensions.codehilite.CodeHiliteExtension(linenums=False)) # deactivate ugly, table-based line numbering. we are going to use glorious css-only linenums

md = app.config['MARKDOWN_CLASS'](
    output_format='xhtml5', # xml-style in order not to fuck too much with <foreignObject> HTML in SVG (see svg.Map for example)
    extensions=[DisplayRenderableExtension()] + extra_extensions
)

md.references.set_loader(magic_markdown_loader)
