import collections
import io
import re
import math
import time
import datetime
import copy
import json
import bson
import numpy
import shapely.geometry

from poobrains import Markup, app, abort, g, session, locked_cached_property

import poobrains.helpers
import poobrains.errors
import poobrains.form
import poobrains.storage
import poobrains.auth
import poobrains.commenting
import poobrains.svg

from . import util, visualization, editor


def load_dataset(handle):

    if handle.startswith('_'): # underscore marks dynamic use of data source in URLs

        name, *args = handle.split('.')
        name = name[1:] # remove leading underscore
        
        if len(name) > 0 and name in datasources:

            source = datasources[name]

            ds = EphemeralDataset(name)
            ds.title = source['title']

            parameters = {}

            if len(args) > len(source['parameters']):
                raise poobrains.errors.ExposedError(f"Too many parameters for data source {name}. Got {len(args)}, expected up to {len(source['parameters'])}.")

            all_params= list(source['parameters'].items())
            for i, arg in enumerate(args):
                param_name, info = all_params[i]
                parameters[param_name] = info['type'](arg) # WARNING: This might be exploitable due to insufficient input sanitation

            ds.data[name] = source['function'](**parameters)

        else:
            raise poobrains.errors.ExposedError(f'Unknown dynamic dataset: {name}')

    else:

        ds = Dataset.load(handle)
        ds.permissions['read'].check(g.user)

    return ds


class EphemeralDataset(poobrains.auth.Protected):

    class Meta:

        modes = collections.OrderedDict([
            ('teaser', 'read'),
            ('inline', 'read'),
            ('json', 'read'),
            ('raw', 'read'),
            ('full', 'read'),
        ])

    title = None # to override Renderable's @property
    description = poobrains.md.MarkdownString()

    def __init__(self, name=None, title=None, description=None, data=None):

        self._name = name
        self.title = title or type(self).__name__
        self.description = description or ''
        self.data = data if not data is None else {}

    def __len__(self):
        # returns number of columns
        return len(self.data)

    def __getitem__(self, column_name):
        return self.data[column_name]

    def __setitem__(self, column_name, column):
        self.data[column_name] = column

    def __delitem__(self, column_name):
        del(self.data[column_name])

    def __iter__(self):
        return self.data.__iter__()
    
    def __contains__(self, item):
        return item in self.data

    @classmethod
    def load(cls, **kwargs):
        raise NotImplementedError(f"Need to implement {cls.__name__}.load.")

    def items(self):
        return self.data.items()

    def values(self):
        return self.data.values()

    def keys(self):
        return self.data.keys()

    def columns_of_dtype(self, dtypes):

        if not isinstance(dtypes, (list, tuple, set)):
            dtypes = [dtypes]

        columns = {}

        unchecked = dict(self.items())
        for dtype in dtypes:
            new = set()
            for column_name, column in unchecked.items():
                if column['dtype'] == dtype:
                    new.add(column_name)

            for name in new:
                columns[name] = unchecked[name]
                del(unchecked[name])

        return columns

    def duplicate_column(self, name, observations=False):

        """
            returns a copy of a columns data, usually without the
            observations – unless the `observations` evaluates to `True`.
            function name might be a slight misnomer.
        """

        copy_keys = {'title', 'description', 'dtype', 'color'}
        new_column = { k:self[name][k] for k in copy_keys}

        if observations:
            new_column['observations'] = copy.deepcopy(self[name]['observations'])
        else:
            new_column['observations'] = {}

        return new_column

    def merge(self, other, joins=None, operation='and', prefix=None):

        """

        Merge data from another dataset into this one.

        Parameters:

        * `other`:      `[Ephemeral]Dataset`, the dataset to merge into this one.
        * `joins`:      `list` of `dict`, definition of joins to "connect" the
                        datasets on. Every one of these dicts has three keys:

                        * `column_self`:    `str`, name of the column in the
                                            *currently edited* dataset that is
                                            being matched on.
                        * `column_other`:   `str`, column in the *other* dataset
                                            that is being matched on.
                        * `tolerance`:      Tolerance of the matching. For
                                            numeric columns, this is a `float`
                                            defining a valid deviation from
                                            strict equality up to which an
                                            observation counts as 'matched'.
                                            For other dtypes, things might be
                                            different.
        * `operation`:  `str`, the boolean operation used to chain individual
                        joins. Can be either `and` or `or`.
        ` `prefix`:     `str|None`: A prefix to be applied to all column names
                        of the *other* dataset before its data is finally being
                        merged into this one.
        """

        if joins is None or len(joins) == 0:

            # dumb merge, all columns from self.other keep all their current index values
            for column_name, column in other.items():

                if prefix:
                    column_name = prefix + column_name

                if column_name in self:
                    continue # skip this column if we would otherwise overwrite an existing column
                else:
                    self[column_name] = column

        else:

            join_matches = []

            for joindef in joins:

                matches = set()

                column_self_name = joindef['column_self']
                column_other_name = joindef['column_other']

                column_self = self[column_self_name]
                column_other = other[column_other_name]

                for index, observation in column_self['observations'].items():

                    indices_other = other.column_indices_by_value(
                        column_other_name,
                        observation,
                        joindef
                    )

                    if len(indices_other) == 1: # clean match
                        matches.add((index, indices_other[0]))

                    else: # uh-oh

                        indices_self = self.column_indices_by_value(
                            column_self_name,
                            observation,
                            joindef
                        )

                        if len(indices_other) == len(indices_self): # means we still got a 1:1 match
                            for match in zip(indices_self, indices_other):
                                matches.add(match)

                        else:
                            poobrains.flash("Ambiguous match in merge. Potential dataloss, buggery or other shitfuckery!", 'error') # FIXME: Should this throw an exception? for sure it shouldn't do a flash, because that breaks CLI
                            app.logger.error(f"Dataset merge: Multiple indices for value {observation}!")
                            app.logger.error(f"Affected column pair: {column_self_name} / {column_other_name}")

                join_matches.append(matches)

            # combine matches found by all joins according to supplied boolean operation
            if operation == 'and':
                combined_matches = join_matches[0].intersection(*join_matches[1:])
            else: # 'or'
                combined_matches = join_matches[0].union(*join_matches[1:])

            # finally, transfer structural definitions of new columns from self.other
            for column_name, column in other.items():

                if prefix:
                    column_name = prefix + column_name

                if column_name not in self:

                    new_column = {
                        'title': column['title'],
                        'description': column['description'],
                        'dtype': column['dtype'],
                        'color': column['color'],
                        'observations': {}
                    }

                    for index_own, index_other in matches:
                        new_column['observations'][index_own] = column['observations'][index_other]

                    self[column_name] = new_column

    #def filter(self, mode, operation, filters):
    #def filter(self, filters):
    def filter(self, filters, operation='and'):

        filter_indices = []

        for filterdef in filters:

            column = self[filterdef['column']] 
            dtype = util.__dtypes__[column['dtype']]

            if filterdef['value_column']:
                filter_value = self[filterdef['value_column']]
            else:
                filter_value = filterdef['value']

            indices = set()
            for index, observation in column['observations'].items():

                # FIXME/TODO: extract a, b, tolerance, load column dtype, call filter_func

                if isinstance(filter_value, dict):
                    try:
                        comparison_value = filter_value['observations'][index]
                    except IndexError: # no corresponding row in value_column
                        continue # skip this row, go to the next one
                else:
                    comparison_value = filter_value

                if dtype.filter_func(observation, comparison_value, filterdef):
                    indices.add(index)

            filter_indices.append(indices)

        if len(filters) == 1:
            matching_indices = filter_indices[0]
        else:
            if operation == 'and':
                matching_indices = filter_indices[0].intersection(*filter_indices[1:])
            else: # 'or'
                matching_indices = filter_indices[0].union(*filter_indices[1:])

        for column_name, old_column in self.items():

            new_column = self.duplicate_column(column_name)
            for index in matching_indices:
                if index in old_column['observations']:
                    new_column['observations'][index] = old_column['observations'][index]

            self[column_name] = new_column

    @property
    def name(self):
        return self._name or type(self).__name__.lower()

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def rowcount(self):
        #returns the number of overall rows in this dataset
        return max([len(self[colname]['observations']) for colname in self])

    @property
    def column_choices(self):
        choices = []

        for column_name, column in self.items():
            choices.append((column_name, f"{column_name}: {column['title']}"))

        return choices

    def column_choices_of_dtype(self, dtypes):

        choices = []

        columns = self.columns_of_dtype(dtypes)
        for column_name, column in columns.items():
            choices.append((column_name, f"{column_name}: {column['title']}"))

        return choices

    @locked_cached_property
    def ref_id(self):
        return f"dataset-{self.name}"

    @locked_cached_property
    def empty(self):
        return 0 == sum([len(column['observations']) for column in self]) # sum length of all columns

    @locked_cached_property
    def plots(self):
        return EphemeralDatasetPlot.select().where(EphemeralDatasetPlot.dataset == self.__class__.__name__.lower())

    def validate(self):

        """ raise ValueError if data is malformed. """

        for name, column in self.items():

            if not isinstance(column, dict):
                raise ValueError(f"Dataset '{self.name}': column '{name}' must be dict, but is {type(column).__name__}.")

            required_keys = {'title', 'description', 'dtype', 'color', 'observations'}
            actual_keys = set()
            for required_key in required_keys:
                if not required_key in column:
                    raise ValueError(f"Dataset '{self.name}': Key '{required_key}' missing in column '{name}'.")
                actual_keys.add(required_key)

            if required_keys != actual_keys:
                invalid_keys = actual_keys - required_keys
                raise ValueError(f"Dataset '{self.name}': Invalid key in column '{name}': {invalid_keys}.")

            if not isinstance(column['observations'], dict):
                raise ValueError(f"Dataset '{self.name}': data[{name}]['observations'] must be dict, but is {type(column['observations']).__name__}.")

            for index, observation in column['observations'].items():

                if not isinstance(observation, util.__dtypes__[column['dtype']].dtype_class):
                    raise ValueError(f"Dataset '{self.name}': Type mismatch in data['{name}']['observations'][{index}] must be {column['dtype']}, but is {type(observation).__name__}.")

    def column_index(self, column_name):
        return [k for k in self[column_name]]

    def column_indices_by_value(self, column_name, value, joindef):

        dtype = util.__dtypes__[self[column_name]['dtype']]
        indices = []
        for index, observation_value in self[column_name]['observations'].items():
            if dtype.match_func(observation_value, value, joindef):
                indices.append(index)

        return indices


    def complete_index(self):

        index = set()
        for column in self.values():
            for idx in column['observations']:
                index.add(idx)

        return list(index) # list implicitly reorders

    @property
    def has_geo(self):
        for column in self.values():
            if column['dtype'] == 'geometry':
                return True
        return False

    @classmethod
    def from_dict(cls, d, whole=False):

        if whole:
            rough_data = d['data']
        else:
            rough_data = d

        data = util.dict_to_data(rough_data)

        if whole:
            ds = cls()
            ds.title = d['title']
            ds.description = poobrains.md.MarkdownString(d['description'])
            ds.data = data

            return ds
        return data

    def to_dict(self, whole=False):

        prepro_data = util.data_to_dict(self.data)
       
        if whole:
            return {
                'title': self.title,
                'description': str(self.description),
                'data': prepro_data,
            }

        return prepro_data

    @classmethod
    def deserialize(cls, raw, format='json'):

        if format == 'json':
            deserializer = json.loads
        elif format == 'bson':
            deserializer = bson.loads
        else:
            raise ValueError(f"Invalid serialization format: {format},  must be 'json' or 'bson'")

        return cls.from_dict(deserializer(raw), whole=True)

    def serialize(self, format='json', whole=False):

        """
            * `whole`: `bool`, whether to serialize the whole dataset or just its `.data`.
        """

        if format == 'json':
            serializer = json.dumps
        elif format == 'bson':
            serializer = bson.dumps
        else:
            raise ValueError(f"Invalid serialization format: {format},  must be 'json' or 'bson'")

        return serializer(self.to_dict(whole=whole))


    def clear(self):
        self.data.clear()

    def render(self, mode='teaser'):

        if mode == 'json':
            return Markup(self.serialize(format='json'))

        return super(EphemeralDataset, self).render(mode=mode)

    @property
    def plottable(self):
        return self.plots.count() > 0

    def plot(self, plot_info=None):

        if plot_info is None:
            if self.plots.count == 0:
                raise ExposedError("This dataset has no associated plot definitions.")
            plot_info = self.plots[0].info

        if not isinstance(plot_info, dict) or not ('kind' in plot_info and 'layers' in plot_info):
            raise ValueError("Invalid plot data.")

        kind = plot_info['kind']

        plot_kinds = visualization.Plot.class_children_keyed()

        if not kind in plot_kinds:
            raise ValueError(f'Unknown plot kind: {kind}')

        return plot_kinds[kind](dataset=self, layers=plot_info['layers'], options=plot_info.get('options'))

    def table(self, max_rows=4):

        """
        Returns a renderable Table object showing the data in this dataset.
        """

        columns = ['index'] + list(self.keys())
        t = poobrains.rendering.Table(title=self.title, columns=columns)

        if max_rows == math.inf:
            source = self.complete_index()
        else:
            source = self.complete_index()[:max_rows]

        for index in source:
            row_data = [index]
            for column in self.values():
                if index in column['observations']:
                    value = poobrains.helpers.truncate_string(str(column['observations'][index]), 100)
                else:
                    value = ""

                row_data.append(value)
            t.append(*row_data)

        return t

    def save(self, name=None, owner=None):

        """
        Convert this EphemeralDataset into a Dataset and save it.
        """

        ds = Dataset()

        #ds.owner = owner or poobrains.auth.User.get(poobrains.auth.User.id == 1)
        ds.owner = owner or g.user
        
        now = int(time.time())
        ds.name = name or poobrains.helpers.clean_string(f"{self.name}-{now}")
        ds.title = f'{self.title}@{str(datetime.datetime.fromtimestamp(now))}'
        ds.description = self.description
        ds.data = self.data
        ds.save(force_insert=True)

        return ds


class DataField(poobrains.storage.fields.BSONField):

    form_widget = None

    def db_preprocess(self, value):
        return util.data_to_dict(value)

    def python_postprocess(self, value):
        return util.dict_to_data(value)


class Dataset(EphemeralDataset, poobrains.commenting.Commentable):

    class Meta:

        modes = collections.OrderedDict([
            ('add', 'create'),
            ('teaser', 'read'),
            ('inline', 'read'),
            ('full', 'read'),
            ('edit', 'update'),
            ('delete', 'delete')
        ])

    title = poobrains.storage.fields.CharField()
    description = poobrains.md.MarkdownField(null=True)
    data = DataField()

    form_edit = editor.DataEditor

    def __init__(self, *args, **kwargs):
        super(EphemeralDataset, self).__init__(*args, **kwargs)
        if self.data is None:
            self.data = {}

    @locked_cached_property
    def ref_id(self):
        return f"dataset-{self.name}"

    @classmethod
    def load(self, handle):
        return super(poobrains.commenting.Commentable, self).load(handle)

    @locked_cached_property
    def plots(self):
        return DatasetPlot.select().where(EphemeralDatasetPlot.dataset == self.__class__.__name__.lower())

    def form(self, mode=None):

        f = super(poobrains.commenting.Commentable, self).form(mode=mode)
        if mode == 'edit':
            del(f['tags'])
        return f

    def save(self, **kwargs):
        return super(EphemeralDataset, self).save(**kwargs)


class EphemeralDatasetPlot(poobrains.storage.Named):

    dataset = poobrains.storage.fields.CharField()
    info = poobrains.storage.fields.BSONField(default={})

class DatasetPlot(EphemeralDatasetPlot):

    dataset = poobrains.storage.fields.ForeignKeyField(Dataset, backref='plots')
